 
open Batteries_uni (* Non-threaded version *)
open Set

(* my modules *)
open Module_KatExpressions      (* a module for kat exps *)
open Module_LinearKatExps       (* a module for kat exps in the linear form *)
open Module_PairsKatExps        (* a module for pairs of kat exps *)
open Module_Utils
open Lexer
open Parser
open LexerAssumptions
open ParserAssumptions

exception ERROR of string

let opt2a = ref true            (* UP2Congr first version *)
let opt2b = ref false           (* UP2Congr second version - compares the two nf concurrently *)
let opts = 
  [ 
    (opt2a, true, "UP2Congr first version");
    (opt2b, false, "UP2Congr second version - compares the two nf concurrently")
  ]

(* Function Aliases *)
let lsRemoveRepet = MUtils.lsRemoveRepetitions
let getAtAndTp = MKE.getAtAndTp

let rec fe1 axiomsEq at t' e1 exp =
  match exp with
      MKE.Prog p      ->  let list =
			    if (List.mem_assoc p axiomsEq) then
			      let gatherPostConds atom p =
				let postconds = [? List: postcond | (p',(precond,postcond)) <- List: axiomsEq ; p'=p ; MKE.evaluate atom t' precond ?] in 
				if postconds=[] then MKE.(Bexp One) else MKE.(Bexp (buildConj postconds)) in
			      let lst1 = [? List: (atom, p, gatherPostConds atom p) | atom <- List: at ; MKE.e atom t' e1 ?] in
			      let lst' = [? List: atom | atom <- List: at ; MKE.e atom t' e1 ; List.for_all (function (atom',x,y) -> if (atom'!=atom) then true else false) lst1 ?] in
			      let lst2 = [? List: (atom,p,MKE.(Bexp One)) | atom <- List: lst' ?]
			      in List.append lst1 lst2
			    else [? List: (atom,p,MKE.(Bexp One)) | atom <- List : at ; MKE.e atom t' e1 ?]
			  in MLinearKE.setLinearKEfromList list
    | MKE.Bexp b        ->  SetLinearKE.empty
    | MKE.Star exp'     ->  MLinearKE.map (MLinearKE.concat3rd exp) (fe1 axiomsEq at t' e1 exp')
    | MKE.Conc (exp1::ls)  ->  let exp2 = MKE.buildConcat ls
			       in let set1 = MLinearKE.map (MLinearKE.concat3rd exp2) (fe1 axiomsEq at t' e1 exp1)
				  in let set2 = fe1 axiomsEq at t' (MKE.Conc [exp1;e1]) exp2
				     in SetLinearKE.union set1 set2
    | MKE.Conc []       ->  raise (ERROR "function fe1 received an unexpected pattern: Conc []")
    | MKE.Or set        ->  MLinearKE.unions (List.map (fe1 axiomsEq at t' e1) (SetKE.elements set))

(* axiomsEq is a list of this form (example): [(1,(b1,b2)); (3,(One,b3)); ...] 
   which reads "the axiomsEq are b1p_1 = b1p_1b2 and p_3 = p_3b3 and ..."     *)
let rec f axiomsEq at t' exp =
  match exp with
      MKE.Prog p  ->  let list =
			if (List.mem_assoc p axiomsEq) then
			  let gatherPostConds atom p =
			    let postconds = [? List: postcond | (p',(precond,postcond)) <- List: axiomsEq ; p'=p ; MKE.evaluate atom t' precond ?] in 
			    if postconds=[] then MKE.(Bexp One) else MKE.(Bexp (buildConj postconds)) in
			  let lst1 = [? List: (atom, p, gatherPostConds atom p) | atom <- List: at ?] in 
			  let lst' = [? List: atom | atom <- List: at ; 
				      List.for_all (function (atom',x,y) -> if (atom'!=atom) then true else false) lst1 ?] in
			  let lst2 = [? List: (atom,p,MKE.(Bexp One)) | atom <- List: lst' ?]
			  in List.append lst1 lst2
			else [? List: (atom,p,MKE.(Bexp One)) | atom <- List : at ?]
		      in MLinearKE.setLinearKEfromList list
    | MKE.Bexp b     ->  SetLinearKE.empty
    | MKE.Star exp'  ->  MLinearKE.map (MLinearKE.concat3rd exp) (f axiomsEq at t' exp')
    | MKE.Conc (exp1::ls)  ->  let exp2 = MKE.buildConcat ls
			       in let set1 = MLinearKE.map (MLinearKE.concat3rd exp2) (f axiomsEq at t' exp1)
				  in let set2 = fe1 axiomsEq at t' exp1 exp2
				     in SetLinearKE.union set1 set2
    | MKE.Conc []   ->  raise (ERROR "function f received an unexpected pattern: Conc []")
    | MKE.Or set    ->  MLinearKE.unions (List.map (f axiomsEq at t') (SetKE.elements set))

let gAux atom prog ls = [? List: exp | (atom',prog',exp) <- List : ls; atom=atom'; prog=prog' ?]
let g axiomsEq at t' exp = 
  let exp_lin = SetLinearKE.elements (f axiomsEq at t' exp) in   
  MLinearKE.setLinearKEfromList [? List : (atom,prog,MKE.buildOr (gAux atom prog exp_lin)) | (atom,prog,_) <- List : exp_lin ?] ;;

let rec der atom prog lst =
match lst with
    (atom',prog',e)::ls  ->  if atom'=atom & prog'=prog then e else der atom prog ls
  | []                   ->  MKE.(Bexp Zero) ;;

let derivatives lf_set1 lf_set2 = 
  let ls_exp1 = SetLinearKE.elements lf_set1 in
  let ls_exp2 = SetLinearKE.elements lf_set2 in
  [? List : (atom, prog, der atom prog ls_exp1, der atom prog ls_exp2) | (atom,prog,_) <- List : (List.append ls_exp1 ls_exp2) ?]

let rec inCongrb e1 e2 rel = 
  let form1 = MKE.NonFinal (SetKE.empty, [e1]) in
  let form2 = MKE.NonFinal (SetKE.empty, [e2]) in
  foo1 (expand form1 rel) (expand form2 rel) rel
and foo1 form1 form2 rel = 
  match (form1,form2) with
    (MKE.Final nf1, MKE.Final nf2)                  ->  SetKE.equal nf1 nf2
  | (MKE.NonFinal (s1,ls1), MKE.Final nf2)          ->  
    if (SetKE.equal nf2 s1) then true else foo1 (expand form1 rel) form2 rel
  | (MKE.Final nf1, MKE.NonFinal (s2,ls2))          ->  
    if (SetKE.equal nf1 s2) then true else foo1 form1 (expand form2 rel) rel
  | (MKE.NonFinal (s1,lsx), MKE.NonFinal (s2,lsy))  ->  
    if (s1=s2) then true else foo1 (expand form1 rel) (expand form2 rel) rel
and expand form rel = 
  match form with
    (MKE.NonFinal (s,h::tl))  ->
      let newSet = SetKE.add h s in
      let newList = 
	match h with
	  | (MKE.Or set) ->  
	    let lst = [? List: x | h1 <- List: (h::(SetKE.fold ff set []));   (* ie, h1 is a subexp of h *)
		       not(MKE.keEq x h); 
		       x <- List: buildLst h1 newSet rel ?] in    (* ie, x is related to h1 in rel *)
	    lsRemoveRepet (tl@lst)
	  | otherwise    ->  let lst = buildLst h newSet rel in 
			     lsRemoveRepet (tl@lst)
      in if newList=[] then (MKE.Final newSet) else (MKE.NonFinal (newSet,newList))
    | _                       ->  raise (ERROR "expand received something other than NonFinal (s,h::tl)")
and buildLst h newSet rel =    
  (* Returns a list of all x s.t. x is related to h in rel and is not in the nf already *)
  List.append 
    [? List: x | (x,h') <- List: (SetPairsKE.elements rel); h'=h; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero)) ?] 
    [? List: x | (h',x) <- List: (SetPairsKE.elements rel); h'=h; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero))?]
and ff x ls = 
  match x with
    (MKE.Or set)  ->  SetKE.fold ff set ls 
  | otherwise     ->  (x::ls)

let rec inCongra e1 e2 rel = 
  let (exp,ls) as pair' = expand (SetKE.empty, [e1]) rel in
  if (ls=[]) then false    (* this means e1 has no correspondence in R *)
  else match (bar e2 pair' rel) with
    None           ->  true
  | Some (s1,ls1)  ->  let (s2,ls2) = expand_completely (SetKE.empty,[e2]) rel in
		       equal s1 s2
and bar e2 pair rel =
  match pair with
      (_,[])     ->  Some pair           (* means they are not equiv and _ is the complete nf of e1 *)
    | otherwise  ->  let pair' = expand pair rel in
		     if member e2 pair' then None      (* means they are equiv *)
		     else bar e2 pair' rel
and expand (s,ls) rel = 
  if ls=[] then (s,[]) else
    let h = List.hd ls in
    let tl = List.tl ls in
    match (h,tl) with
      | ((MKE.Or set),_) ->  let newSet = SetKE.add h s in
			     let lst = 
			       [? List: x | h1 <- List: (h::(SetKE.fold ff set [])) (* ie,h1 is a subexp of h *)
			       ; not(MKE.keEq x h) 
			       ; x <- List: buildLst h1 newSet rel ?] in  (* ie, x is related to h1 in rel *)
			     let newList = lsRemoveRepet (tl@lst) in
			     (newSet, newList)
      | (h,tl)           ->  let newSet = SetKE.add h s in
			     let lst = buildLst h newSet rel in
			     let newList = lsRemoveRepet (tl@lst) in
			     (newSet, newList)
and buildLst h newSet rel =    
  (* returns a list of all x s.t. x is related to h in rel and is not in the nf already *)
  List.append 
    [? List : x | (x,h') <- List : (SetPairsKE.elements rel) ; h'=h ; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero)) ?] 
    [? List : x | (h',x) <- List : (SetPairsKE.elements rel) ; h'=h ; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero))?]
and ff x ls = 
  match x with
    (MKE.Or set)   ->   SetKE.fold ff set ls 
  | otherwise      ->   (x::ls)
and expand_completely pair rel = 
  let (e',ls') as pair' = expand pair rel in 
  if ls'=[] then pair' 
  else expand_completely pair' rel
and member e2 (nf_set, nf_lst) =
  match nf_lst with
    (h::tl)  ->  
      if MKE.keEq e2 (MKE.buildOr (SetKE.elements (SetKE.add h nf_set))) then true else member e2 (nf_set,tl)
  | []       ->  false 
and equal nf1 nf2 = MKE.keEq (MKE.buildOr (SetKE.elements nf1)) (MKE.buildOr (SetKE.elements nf2))

let rec equivP axiomsEq t' at set h = 
  if (SetPairsKE.is_empty set) then (true, h)
  else 
    let (exp1,exp2) = (SetPairsKE.min_elt set) in
    let (at,t') = getAtAndTp exp1 exp2 in
    if (MKE.eAll ~at:at ~t':t' exp1 exp2)=false then 
      let () = if !debugging then print_endline ("The error was in " ^ (MKE.katExp2String exp1) ^ "and" ^ (MKE.katExp2String exp2)) in
      (false, SetPairsKE.add (exp1,exp2) h)
    else 
      let h' = SetPairsKE.add (exp1,exp2) h in
      let s' =
	let sets = (SetPairsKE.union h' (SetPairsKE.remove (exp1,exp2) set)) in
	if !opt2a then
	  MPairsKE.fromList 
	    [? List : (der1,der2) | (_,_,der1,der2) <- List:(derivatives (g axiomsEq at t' exp1) (g axiomsEq at t' exp2)); not (SetPairsKE.mem (der1,der2) sets) ; not(inCongra der1 der2 sets) ?]
	else if !opt2b then
	  MPairsKE.fromList 
	    [? List : (der1,der2) | (_,_,der1,der2) <- List:(derivatives (g axiomsEq at t' exp1) (g axiomsEq at t' exp2)); not (SetPairsKE.mem (der1,der2) sets) ; not(inCongrb der1 der2 sets) ?]
	else 
	  MPairsKE.fromList 
	    [? List : (der1,der2) | (_,_,der1,der2) <- List:(derivatives (g axiomsEq at t' exp1) (g axiomsEq at t' exp2)); not (SetPairsKE.mem (der1,der2) sets) ?]  in
      let s' = SetPairsKE.union (SetPairsKE.remove (exp1,exp2) set) s' in
      equivP axiomsEq t' at s' h'

let parse str =
  let linebuf = Lexing.from_string (str ^ "\nEOF") in
  List.hd (Parser.data Lexer.token linebuf)

let parseAxioms str = 
  let linebuf = Lexing.from_string (str ^ "\nEOF") in
  ParserAssumptions.data LexerAssumptions.token linebuf

let rec filterAtAxioms t' at axiomsLeq =
  match axiomsLeq with  (* (b,c) means c <= b *)
      [(b,c)]    ->   List.filter (genPred t' b c) at
    | (b,c)::ls  ->   filterAtAxioms t' (List.filter (genPred t' b c) at) ls
    | []         ->   at
and genPred t' b c atom = 
  (MKE.evaluate atom t' c = false) || (MKE.evaluate atom t' c = true && MKE.evaluate atom t' b = true)

let main (axiomsEq,axiomsLeq) exp1 exp2 = 
  let exp1 = MKE.flattenConcats exp1 in
  let exp2 = MKE.flattenConcats exp2 in
  let (at,t') = MKE.getAtAndTp exp1 exp2 in
  let atAxioms = filterAtAxioms t' at axiomsLeq in
  equivP axiomsEq t' atAxioms (SetPairsKE.singleton (exp1,exp2)) SetPairsKE.empty
    
