{
  open ParserAssumptions
  exception Eof
  exception Error of string
}

rule line = parse
    ([^'\n']* '\n') as line  {line}    (* a line is defined by 0 or more *)
  | eof	{exit 0}		       (* repetitions of any chars other than *)
                                       (* '\n' and terminated by a '\n') *)
and token = parse
  [' ' '\t'] {token lexbuf}        (* skip white spaces and read what follows *)
  | '\n' {EOL}
  | "t" {B}
  | "t_" {B}
  | "p" {P}
  | "p_" {P}
  | "not" {NOT}
  | "¬" {NOT}
  | "~" {NOT}
  | "=" {EQUAL}
  | "<=" {LEQ}
  | "&" {AND}
  | '+' {PLUS}
  | "One" {ONE}
  | "Zero" {ZERO}
  | ['0'-'9']+ as i {INT (int_of_string i)}
  | '(' {LPAREN}
  | ')' {RPAREN}
  | eof {raise Eof}
  | _ { raise (Error (Printf.sprintf "At offset %d: unexpected character.\n" (Lexing.lexeme_start lexbuf))) }


