
open Module_SetBuilder
open Module_NFAGS

exception Error of string

module CallNFAGSEquivChecker : sig
  type automaton = NFAGS.automaton
  val call: automaton -> automaton -> bool*(SetPairsIntSet.t option)*(SetPairsIntSet.t option)
  val opts: ((bool ref)*bool*string) list
  val optIncl1: bool ref
  val optIncl2: bool ref
end
= struct
  type automaton = NFAGS.automaton
      
  (* Function Aliases *)
  let unionR = SetPairsIntSet.union 

  let optIncl1 = ref false
  let optIncl2 = ref true
  let opts = 
    [
      (optIncl1, false, "Use lang. inclusion (incl1) x 2 instead of lang. equality directly.");
      (optIncl2, true, "Use lang. inclusion (incl2) x 2 instead of lang. equality directly.")
    ]

  let call a1 a2 = 
    match (a1,a2) with
      (NFAGS.TransProd x, NFAGS.TransProd y) ->  
	if !optIncl1 then 
	  let (answer1,r1) = InclusionCheckerAGS1.main x y in
	  let (answer2,r2) = InclusionCheckerAGS1.main y x in
	  (answer1&&answer2, Some r1, Some r2)
	else if !optIncl2 then
	  let (answer1,r1) = InclusionCheckerAGS2.main x y in
	  let (answer2,r2) = InclusionCheckerAGS2.main y x in
	  (answer1&&answer2, Some r1, Some r2)
	else
	  let (answer, r1) = EquivalenceCheckerAGS.main x y in
	  (answer, Some r1, None)
    | _                                      ->  raise (Error "Unexpected pattern in CallNFAGSEquivChecker.call.")
      
end
  
