type position = int
exception ERROR of string
exception Found_Zero
exception Found_Zero' of position

open Module_KatExpressions
open Module_KatExpressionsPos
open ParserPolish
open Module_SetBuilder

let print str = print_endline str ; flush stdout
let print_el str = print_string str ; flush stdout

(* the difference between MKEPos2 and MKEPos is that in this module we 
   also distinguish the test symbols, and not only the action ones *)
module rec MKEPos2 : sig 
  type position = int
  type index = int
  type be = MKE.be (* Test of int
	    | One
	    | Zero
	    | NegBexp of be
	    | Disj of SetBexpsPos.t
	    | Conj of SetBexpsPos.t *)
  type ke = Prog of index*position
	    | Bexp of be*position
	    | Star of ke
	    | Conc of ke list
	    | Or of SetKEPos2.t 
  val buildConcat : ke list -> ke
  val katExp2String: ke -> string
  val printKE_nnl: ke -> unit
  val markPositionsInSymbols : MKEPos.ke -> int -> ke
end 
  = struct
    type position = int
    type be = MKE.be (* Test of int
	      | One
	      | Zero
	      | NegBexp of be
	      | Disj of SetBexpsPos.t
	      | Conj of SetBexpsPos.t*)
    type ke = Prog of int*position
	      | Bexp of be*position
	      | Star of ke
	      | Conc of ke list
	      | Or of SetKEPos2.t 
    type index = int

    let buildConcat ls =
      let simplify = true in   (* turns to 0 all concats of the form x..0..y *)
      match ls with
	  [exp]      ->  exp
	| otherwise  ->  
	  if simplify then
	    let rec loop ls =
	      match ls with
		| (Bexp (MKE.Zero,pos))::tl  ->  raise (Found_Zero' pos)
		| [Bexp (MKE.One,_)]         ->  ls
		| (Bexp (MKE.One,_))::tl     ->  loop tl
		| (Bexp (r,pos))::tl     ->  
	  (*      if isAlwaysFalse r then      Commented since it's probably too heavy to compute..
		    raise Found_Zero
		  else  *)
		  (Bexp (r,pos))::(loop tl)
		| exp::tl               ->  exp::(loop tl)
		| []                    ->  []
	    in match (try loop ls with (Found_Zero' pos) -> [Bexp (MKE.Zero,pos)]) with
	      [h]    ->  h
	    | h::tl  ->  Conc (h::tl)
	    | []     ->  raise (ERROR "MKEPos2.buildConcat received []")
	  else (Conc ls)

    let rec buildOr lst =
      match lst with
	  []         ->  raise (ERROR "MKEPos2.buildOr received [] as argument ")
	| [l]        ->  l
	| otherwise  ->  
	  let rec aux lst' =
	    match lst' with
	      (Or set)::ls    ->  let set' = aux ls in
				  (SetKEPos2.union set set')
	    | [Bexp (MKE.Zero,pos)]     ->  SetKEPos2.singleton (Bexp (MKE.Zero,pos))
	    | (Bexp (MKE.Zero,_))::ls   ->  aux ls
	    | r::ls           ->  let set' = aux ls in
				  (SetKEPos2.union (SetKEPos2.singleton r) set')
	    | []              ->  SetKEPos2.empty
	  in let newSet = aux lst in
	     buildOr_set newSet
    and buildOr_set set = 
      match (SetKEPos2.cardinal set) with
	1  ->  SetKEPos2.min_elt set
      | 0  ->  raise (ERROR "MKEPos2.buildOr_set received an empty set")
      | n  ->  Or set
	
    (* parser helper functions start here *)

    let rec setBexpsfromList lst =
      match lst with
	  []       -> SetBexpsPos.empty
	| l::ls    -> SetBexpsPos.add l (setBexpsfromList ls)
	       
(*    let rec unions lst =
      match lst with
	  []	 ->  SetKEPos.empty
	| l::ls	 ->  SetKEPos.union l (unions ls) 

    let countTestsInKe exp =
      let rec aux exp base = 
	match exp with
	    Bexp b         ->  MKE.countTestsInBexp b + base
	  | Prog _         ->  base
	  | Star e         ->  aux e base
	  | Conc ls        ->  List.fold_right aux ls base
	  | Or set         ->  SetKEPos.fold aux set base
      in aux exp 0

    let map f set = SetKEPos.fold (fun x y -> SetKEPos.add (f x) y) set SetKEPos.empty
	    
(*    let parse str =
      let linebuf = Lexing.from_string (str ^ "\nEOF") in
      List.hd (Parser.data Lexer.token linebuf) *)
*)

    (* Finds the next (pos or neg) action position that should be used in the given expression *)
    let findNextActPosition exp i = 
      let rec foo exp : (position list) = 
	match exp with
	  MKEPos.Prog (_,pos)  ->  [pos]
	| MKEPos.Star e        ->  foo e
	| MKEPos.Bexp _        ->  []
	| MKEPos.Conc ls       ->  List.flatten (List.map foo ls)
	| MKEPos.Or set        ->  List.flatten (List.map foo (SetKEPos.elements set))  in
      let ls = List.fast_sort compare (foo exp) in
      if ls=[] then
	if i=1 then 2 else -2
      else
	let min = List.hd ls in
	if min < 0 then (min-1)
	else ((List.hd (List.rev ls))+1)

    let rec katExp2String exp =
      match exp with
	Prog (p,i) ->  "(p"^(string_of_int p)^","^(string_of_int i)^")"
      | Bexp (b,i) ->  "("^MKE.bexp2String b^","^(string_of_int i)^")"
      | Star exp   ->  "(" ^ (katExp2String exp) ^ ")*"
      | Conc lst   ->  
	let rec aux = function  
	[l]    ->  katExp2String l
	  | l::ls  ->  (katExp2String l) ^ (aux ls)
	  | []     ->  raise (ERROR "aux in katExp2String.Conc received []")
	in "(" ^ (aux lst) ^ ")"
      | Or set     ->  let rec printAux = function
    []     -> ""
	| l::ls  -> "+" ^ (katExp2String l) ^ (printAux ls)
		       in let set_elements = SetKEPos2.elements set
			  in if (set_elements == []) then (raise (ERROR "function printAux in MKE.katExp2String received an empty set")) else "(" ^ (katExp2String (List.hd set_elements)) ^ (printAux (List.tl set_elements)) ^ ")"
			      
    let printKE_nnl exp = print_str (katExp2String exp)


    let incPosition (pos:position ref) = 
      if !pos>0 then pos:=(!pos)+1 else pos:=(!pos)-1

    (* Each element of Sigma U B (where B is the free BA) is marked with its position;
       It is not necessarily the same position in which it occured in the input expression *)
    let rec markPositionsInSymbols (exp:MKEPos.ke) (i:int) : ke = 
      loop (ref (findNextActPosition exp i)) exp
    and loop (pos:position ref) (exp:MKEPos.ke) = 
      match exp with 
	MKEPos.Prog (p,posi) ->  MKEPos2.Prog (p,posi)
      | MKEPos.Star e        ->  MKEPos2.Star (loop pos e)
      | MKEPos.Bexp b        ->  let posl = !pos in (incPosition pos; MKEPos2.Bexp (b,posl))
      | MKEPos.Or set        ->  
	MKEPos2.Or (SetKEPos.fold (fun a b -> SetKEPos2.add (loop pos a) b) set SetKEPos2.empty)
      | MKEPos.Conc ls       ->  MKEPos2.Conc (List.map (loop pos) ls)
	    
  end
and OrderKEPos : Set.OrderedType   (* an order for expressions inside an 'Or' *)
  = struct
    type t = MKEPos2.ke      
    let compare = Pervasives.compare
  end
and SetKEPos2 : Set.S with type elt = MKEPos2.ke   (* a set for expressions inside an 'Or' *)
  = Set.Make( OrderKEPos )
		      and OrderBexpsPos : Set.OrderedType
    = struct
      type t = MKEPos2.be
      let compare = Pervasives.compare
    end
		     and SetBexpsPos : Set.S with type elt = MKEPos2.be
  = Set.Make( OrderBexpsPos )
  
