
open Module_KatExpressions
open Module_KatExpressionsPos
open Module_Utils
open Lexer
open Parser

module rec MPairsActBe : sig 
  type t = (int*int) * MKE.be
  val buildSetFromList : t list -> SetPairsActBe.t ;;
  val isRedundant : SetPairsActBe.t -> bool ;;
  val printSet : SetPairsActBe.t -> unit ;;
  val setPairsActBe2String : SetPairsActBe.t -> string ;;
  val map : (t -> t) -> SetPairsActBe.t -> SetPairsActBe.t ;;
  val unions : SetPairsActBe.t list -> SetPairsActBe.t;;
end 
  = struct

    type t = (int*int) * MKE.be
	
    let rec buildSetFromList lst =
      match lst with
	  []       ->  SetPairsActBe.empty
	| l::ls    ->  SetPairsActBe.add l (buildSetFromList ls) ;;
    
    let isRedundant set =
      SetPairsActBe.for_all (fun (_,be) -> MKE.isAlwaysFalse be) set

    let setPairsActBe2String set =
      let rec loop set =
	if (SetPairsActBe.is_empty set) then "<this set is empty>" else
	  if (SetPairsActBe.cardinal set)=1 then
	    let ((x,z),y) = SetPairsActBe.min_elt set in
	    "(" ^ (MKEPos.katExp2String (MKEPos.Prog (x,z))) ^ "," ^ (MKE.bexp2String y) ^")"
	  else 
	    let ((x,z),y) = SetPairsActBe.min_elt set in
	    "(" ^ (MKEPos.katExp2String (MKEPos.Prog (x,z))) ^ "," ^ (MKE.bexp2String y) ^"),\n" ^ (loop (SetPairsActBe.remove ((x,z),y) set)) in
      "{" ^ (loop set) ^ "}"

    let printSet set = print_endline (setPairsActBe2String set)
  
    let map f set = SetPairsActBe.fold (fun x y -> SetPairsActBe.add (f x) y) set SetPairsActBe.empty

    let rec unions lst =
      match lst with
	[]	 ->  SetPairsActBe.empty
      | l::ls	 ->  SetPairsActBe.union l (unions ls)

  end
and OrderPairsActBe : Set.OrderedType
  = struct
    type t = MPairsActBe.t    
    let compare = Pervasives.compare
  end 
and SetPairsActBe : Set.S with type elt = MPairsActBe.t
  = Set.Make( OrderPairsActBe )
