
(* the functor SO creates an Order (pre-requisite for the creation of any set) *)
module SO (M: sig type t end)
  = struct
    type t = M.t
    let compare = Pervasives.compare
  end
  
(* the functor S creates a Set of type 2^t *)
module S (P: sig type t end): Set.S with type elt = P.t
  = Set.Make (SO (P))

(* a module with useful functions over sets *)
module SM (M: sig type t end): 
sig
  type t = M.t
  module S: Set.S with type elt = t and type t = S(M).t
  type set = S.t
  val fromList2Set: t list -> set
end
= struct
  type t = M.t
  module S = S (M)
  type set = S.t
  let fromList2Set ls = List.fold_left (fun a b -> S.add b a) S.empty ls
end


(* the functor OR creates an Order for a Relation *)
module OR (M: sig type t end)
  = struct
    type s = S(M).t
    type t = s*s
    let compare = Pervasives.compare
  end

(* the functor R creates a Relation of type 2^ *)
module R (P:sig type t end): Set.S with type elt = S(P).t*S(P).t
  = Set.Make (OR (P))

(* a module with useful functions over relations *)
module RM (P:sig type t end):
sig 
  type t = P.t
  module S: Set.S with type elt = P.t and type t = S(P).t
  module R: Set.S with type elt = S.t*S.t and type t = R(P).t
end
= struct
  type t = P.t
  module R = R (P)
  module S = S (P)
end

(* instantiation of a Set of ints *)
module T: sig 
  type t=int 
  val fromElt2String: t -> string 
end 
  = struct 
    type t=int 
    let fromElt2String = string_of_int
  end
module ModInt = SM (T)
module SetInt = ModInt.S

(* instantiation of a Relation of sets of ints *)
module ModIntRel = RM (T)
module SetPairsIntSet = ModIntRel.R
