(* To use this module for a Map with a different key-to-content type, simply modify 
   'myKey' and 'ontent' and change the name of the module after. *)
type myKey = string
type myContent = float
let print str = print_endline str; flush stdout
let print_str str = print_string str; flush stdout
  
module rec MString2FloatMap: sig
  type key = MyMap.key
  type content = myContent
  type map = content MyMap.t
  val sum: map -> map -> map
  val divideByN: int -> map -> map
  val empty: map
  val fromContent2String: content -> string
  val fromKey2String: key -> string
  val fromMap2String: map -> string
  val increment: key -> content -> map -> map
  val map: (content -> 'b) -> map -> 'b MyMap.t
  val printMap: map -> unit
  val printMap_nl: map -> unit
end
  = struct
      
    type key = MyMap.key
    type content = myContent
    type map = content MyMap.t

    let divideByN n map = MyMap.map (fun m -> m/.(float n)) map

    let empty = MyMap.empty

    let fromContent2String content = string_of_float content

    let fromKey2String key = key 

    let fromMap2String map = 
      "[ " ^ 
	(MyMap.fold (fun key cont str -> (fromKey2String key)^": "^(fromContent2String cont)^" "^str) map "")
      ^ "]"

    (* If the key 'key' already exists in map, its correspondence will be incremented in 'increment'; 
       otherwise, a new entry is created in the map *)
    let increment key increment map = 
      let new_value = 
	try ((MyMap.find key map) +. increment) with Not_found -> increment in
      MyMap.add key new_value map

    (* Sums two maps; pre-requisite: the two maps must have exactly the same keys *)
    let sum map1 map2 =
      MyMap.mapi (fun key a -> (MyMap.find key map1) +. a) map2

    let map = MyMap.map
	
    let printMap map = print_str (fromMap2String map)
    let printMap_nl map = print (fromMap2String map)
      
  end
and MapOrder : Map.OrderedType 
  = struct
    type t = myKey
    let compare = Pervasives.compare
  end
and MyMap : Map.S with type key = myKey
  = Map.Make (MapOrder)
