%{

  (* OCamls imports and others go here *)
  open Module_KatExpressions
  exception ERROR of string
      
  (* aliases *)
  let one = MKE.(Bexp MKE.One)
  let test i = MKE.(Bexp (MKE.Test i))
  let zero = MKE.(Bexp MKE.Zero)
%}

%token P
%token B
%token <int> INT
%token NOT
%token ONE ZERO
%token PLUS STAR
%token LPAREN RPAREN	(* left and right parentheses *)
%token EOL		(* end of the line *)

%start data
%type <MKE.ke list> data

%%

data:
  r=exp ; EOL {[r]}

exp:
  t=term ; PLUS ; ts=plus {if t=MKE.(Bexp Zero) then MKE.buildPlus ts else MKE.buildPlus (t::ts)}
| t=term {t} 

plus:
  t=term ; PLUS ; ts=plus {if t=MKE.(Bexp Zero) then ts else t::ts}
| t=term {[t]}

term:
| f=factor ; fs=concat {if f=MKE.(Bexp Zero) then MKE.(Bexp Zero) else if fs=[MKE.(Bexp Zero)] then MKE.(Bexp Zero) else if fs=[MKE.(Bexp One)] then f else if f=MKE.(Bexp One) then MKE.buildTimes fs else MKE.buildTimes (f::fs)}
| f=factor {f}

concat:
| f=factor ; fs=concat {if f=MKE.(Bexp Zero) then [MKE.(Bexp Zero)] else if f=MKE.(Bexp One) then fs else if fs=[MKE.(Bexp Zero)] then [MKE.(Bexp Zero)] else if fs=[MKE.(Bexp One)] then [f] else f::fs}
| f=factor {[f]}

factor:
| p=symb {p}
| s=symb ; STAR {if s=MKE.(Bexp Zero) then MKE.(Bexp One) else if s=MKE.(Bexp One) then MKE.(Bexp One) else MKE.Star s}
                
bool:
  ONE {one}
| ZERO {zero}
| B ; i=INT {test i}
| NOT ; b=bool {match b with MKE.Bexp f -> MKE.Bexp (MKE.buildNeg f) | otherwise -> raise (ERROR "The negation operator can only be applied to boolean expressions")}
                                                                           
symb:
| P ; i=INT {MKE.(Prog i)}
| b=bool {b}
| LPAREN e = exp RPAREN {e}
| NOT ; LPAREN e = exp RPAREN {match e with MKE.Bexp f -> MKE.Bexp (MKE.buildNeg f) | otherwise -> raise (ERROR "The negation operator can only be applied to boolean expressions")}
