%{

(* OCamls imports and others go here *)
open Module_KatExpressions
exception ERROR of string
let errMsg_commutCond = "You inserted an expression that appears to be a commutativity condition but it is malformed. For instructions, exit and type ./runme -help.";;
let errMsg_hoareTriple = "You inserted an expression that appears to be an Hoare triple but it is malformed. For instructions, exit and type ./runme -help.";;
%}

%token P
%token B
%token <int> INT
%token EQUAL
%token AND
%token NOT
%token ONE ZERO
%token PLUS
%token LEQ
%token LPAREN RPAREN	(* left and right parentheses *)
%token EOL		(* end of the line *)

%start data
%type < ( ((int * (MKE.be * MKE.be)) list) * ((MKE.be * MKE.be) list) )> data

%%

data:
  l=line ; EOL {l}

line:
  a=axiomEq ; AND ; ls=line  { match ls with (ls1,ls2) -> (a::ls1,ls2) }
| a=axiomLeq ; AND ; ls=line  { match ls with (ls1,ls2) -> (ls1,a::ls2) }
| a=axiomEq {([a],[])}
| a=axiomLeq {([],[a])}

axiomEq:
  b0=bexp ; P ; i=INT ; EQUAL ; b1=bexp ; P ; j=INT ; b2=bexp {if (not (b0=b1) || (not (i=j))) then raise (ERROR errMsg_hoareTriple) else (i,(b0,b2))}
| P ; i=INT ; EQUAL ; P ; j=INT ; b2=bexp {if (not (i=j)) then raise (ERROR errMsg_hoareTriple) else (i,(MKE.One,b2))}
| b0=bexp ; P ; i=INT ; b2=bexp ; EQUAL ; b1=bexp ; P ; j=INT {if (not (b0=b1) || (not (i=j))) then raise (ERROR errMsg_hoareTriple) else (i,(b0,b2))}
| P ; i=INT ; b2=bexp ; EQUAL ; P ; j=INT {if (not (i=j)) then raise (ERROR errMsg_hoareTriple) else (i,(MKE.One,b2))}
| P ; i=INT ; b1=bexp ; EQUAL ; b2=bexp ; P ; j=INT {if (not (i=j) || not (b1=b2)) then raise (ERROR errMsg_commutCond) else (i,(b1,b1))}
| b1=bexp ; P ; i=INT ; EQUAL ; P ; j=INT ; b2=bexp  {if (not (i=j) || not (b1=b2)) then raise (ERROR errMsg_commutCond) else (i,(b1,b1))}

axiomLeq:
  b0=bexp ; LEQ ; b1=bexp {(b1,b0)}

bexp:
  t=term ; PLUS ; ts=plus {if t=MKE.Zero then MKE.buildDisj ts else MKE.buildDisj (t::ts)}
| t=term {t} 

plus:
  t=term ; PLUS ; ts=plus {if t=MKE.Zero then ts else t::ts}
| t=term {[t]}

term:
| s=symb ; fs=concat {if s=MKE.Zero then MKE.Zero else if fs=[MKE.Zero] then MKE.Zero else if fs=[MKE.One] then s else if s=MKE.One then MKE.buildConj fs else MKE.buildConj (s::fs)}
| s=symb {s}

concat:
| s=symb ; fs=concat {if s=MKE.Zero then [MKE.Zero] else if s=MKE.One then fs else if fs=[MKE.Zero] then [MKE.Zero] else s::fs}
| s=symb {[s]}
                                                                              
symb:
  ONE {MKE.One}
| ZERO {MKE.Zero}
| B ; i=INT {MKE.(Test i)}
| NOT ; B ; i=INT {MKE.(NegBexp (Test i))}
| LPAREN e = bexp RPAREN {e}
| NOT ; LPAREN e = bexp RPAREN {MKE.buildNeg e}
