%{

(* A Parser that distinguishes all action symbols (pi).
   This parser accepts expressions written in Polish notation. *)

(* OCamls imports and others go here *)
open Module_KatExpressionsPos
open Module_KatExpressions
exception ERROR of string

(* This parser has some optimizations available, to turn them on
   change the optimizations value to ON below *)
let optimizations = "OFF"

let treatTimes e1 e2 = 
  if optimizations="ON" then
    (if (e1=MKEPos.(Bexp MKE.Zero) || e2=MKEPos.(Bexp MKE.Zero)) then 
      MKEPos.(Bexp MKE.Zero)
    else if e1=MKEPos.(Bexp MKE.One) then 
      e2
    else if e2=MKEPos.(Bexp MKE.One) then 
      e1 
    else MKEPos.buildTimes [e1;e2])
  else MKEPos.buildTimes [e1;e2]

let treatPlus e1 e2 = 
  if optimizations="ON" then
    if e1=MKEPos.(Bexp MKE.Zero) then e2 else MKEPos.buildPlus [e1;e2]
  else MKEPos.buildPlus [e1;e2]

let rec isBoolean exp = 
  match exp with 
      MKEPos.Bexp _    ->  true
    | MKEPos.Star exp' ->  isBoolean exp'
    | _                ->  false

(* Checks if e is boolean and if so returns 1 *)
let treatStar e = 
  if optimizations="ON" then
    if isBoolean e then MKEPos.(Bexp MKE.One)
    else MKEPos.Star e
  else MKEPos.Star e

let z = ref 1

let incZ () = if !z>0 then z:=(!z)+1 else z:=(!z)-1
let resetZ () = if !z>0 then z:=(-1) else z:=1

let prog i = incZ () ; MKEPos.Prog (i,!z)

let eol r = resetZ () ; [r]

%}

%token P
%token B
%token <int> INT
%token NOT
%token ONE ZERO
%token PLUS STAR
%token TIMES
%token EOL		(* end of the line *)

%start data
%type <MKEPos.ke list> data

%%

data:
  r=exp ; EOL {eol r}

exp:
| PLUS  ; e1=exp ; e2=exp {treatPlus e1 e2}
| TIMES ; e1=exp ; e2=exp {treatTimes e1 e2}
| STAR  ; e=exp {treatStar e}
| ONE {MKEPos.(Bexp MKE.One)}
| ZERO {MKEPos.(Bexp MKE.Zero)}
| P ; i=INT {prog i}
| B ; i=INT {MKEPos.(Bexp (MKE.Test i))}
| NOT ; e = exp {match e with MKEPos.Bexp f -> MKEPos.Bexp (MKE.buildNeg f) | otherwise -> raise (ERROR "The negation operator can only be applied to boolean expressions")}

