%{

  (* A parser for infix notation that distinguishes all action symbols (pi). *)

  (* OCamls imports and others go here *)
  open Module_KatExpressionsPos
  open Module_KatExpressions
  exception ERROR of string

  (* Aliases *)
  let be e = MKE.Bexp e
  let buildNeg f = MKEPos.Bexp (MKE.buildNeg f)
  let buildPlus = MKEPos.buildPlus
  let buildTimes = MKEPos.buildTimes
  let one = MKEPos.(Bexp MKE.One)
  let test i = MKEPos.(Bexp (MKE.Test i))
  let negTest i = MKEPos.(Bexp (MKE.NegBexp (MKE.Test i)))
  let star s = MKEPos.Star s
  let zero = MKEPos.(Bexp MKE.Zero)

  let z = ref 1

  let incZ () = if !z>0 then z:=(!z)+1 else z:=(!z)-1
  let resetZ () = if !z>0 then z:=(-1) else z:=1

  let prog i = incZ () ; MKEPos.Prog (i,!z)

  let eol r = resetZ () ; [r]

%}

%token P
%token B
%token <int> INT
%token NOT
%token ONE ZERO
%token PLUS STAR
%token LPAREN RPAREN	(* left and right parentheses *)
%token EOL		(* end of the line *)

%start data
%type <MKEPos.ke list> data

%%

data:
  r=exp ; EOL {eol r}

exp:
  t=term ; PLUS ; ts=plus {if t=zero then buildPlus ts else buildPlus (t::ts)}
| t=term {t} 

plus:
  t=term ; PLUS ; ts=plus {if t=zero then ts else t::ts}
| t=term {[t]}

term:
| f=factor ; fs=concat {if f=zero then zero else if fs=[zero] then zero else if fs=[one] then f else if f=one then buildTimes fs else buildTimes (f::fs)}
| f=factor {f}

concat:
| f=factor ; fs=concat {if f=zero then [zero] else if f=one then fs else if fs=[zero] then [zero] else if fs=[one] then [f] else f::fs}
| f=factor {[f]}

factor:
| p=symb {p}
| s=symb ; STAR {if s=zero then one else if s=one then one else star s}
     
bool:
  ONE {one}
| ZERO {zero}
| B ; i=INT {test i}
| NOT ; b=bool {match b with MKEPos.Bexp f -> MKEPos.Bexp (MKE.buildNeg f) | otherwise -> raise (ERROR "The negation operator can only be applied to boolean expressions")}

symb:
  P ; i=INT {prog i}
| b=bool {b}
| LPAREN e = exp RPAREN {e}
| NOT ; LPAREN e = exp RPAREN {match e with MKEPos.Bexp f -> buildNeg f | otherwise -> raise (ERROR "The negation operator can only be applied to boolean expressions")}
                               
